#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <cinttypes>

#include "utils.h"

using namespace std;


bin_tree *build_tree(int *len, int *code, int elements) {
	int i,j;
	unsigned x;
	bin_tree *t,*ret=new bin_tree();

	for (i=0;i<elements;i++) {
		t=ret;
		x=1;
		x=1<<(len[i]-1);
		for (j=0;j<len[i];j++) {
			if ((x & code[i])!=0) {
				if (t->bit_1==0) {
					t->bit_1=new bin_tree();
				}
				t=t->bit_1;
			} else {
				if (t->bit_0==0) {
					t->bit_0=new bin_tree();
				}
				t=t->bit_0;
			}

			x=x>>1;
		}
		t->data=i;
	}
	return ret;
}

int get_code(bin_stream *b, bin_tree *tree) {
	bin_tree *t;

	t=tree;
	while (1) {
		if ((b->peek()&1)==1) {
			if (t->bit_1) {
				t=t->bit_1;
				b->read(1);
			} else {
				return t->data;
			}
		} else {
			if (t->bit_0) {
				t=t->bit_0;
				b->read(1);
			} else {
				return t->data;
			}
		}
	}

	return -1; // never used
}

void decode_len(bin_stream *b, bin_tree *tree, int *len, int elements) {
	int i,l,extra,code;

	i=0;
	while (i<elements) {
		l=get_code(b,tree);
		switch (l) {
			case 16:
				extra=b->read(2) + 3;
				if (i) {
					code=len[i-1];
				} else {
					code=0;
				}
				while (extra>0) {
					len[i]=code;
					i++;
					extra--;
				}
				break;
			case 17:
				extra=b->read(3) + 3;
				while (extra>0) {
					len[i++]=0;
					extra--;
				}
				break;
			case 18:
				extra=b->read(7) + 11;
				while (extra>0) {
					len[i++]=0;
					extra--;
				}
				break;

			default:
				len[i++]=l;
				break;
		}

	}
}

void create_code_table(int *len, int *data, int max_bits, int elements) {
	int i;
	int code;
	int bl_count[16];
	int next_code[16];

	for (i=0;i<16;i++) {
		bl_count[i]=0;
	}
	for (i=0;i<elements;i++) {
		bl_count[len[i]]++;
	}

	code=0;
	for (i=1;i<=max_bits;i++) {
		code=(code+bl_count[i-1])<<1;
		next_code[i]=code;
	}
	for (i=0;i<elements;i++) {
		if (len[i]) {
			data[i]=next_code[len[i]]++;
		}
	}
}

int create_dynamic_trees(bin_stream *b, bin_tree **ret_lit, bin_tree **ret_dist) {
	int i;
	int hclen, hdist, hlit;


	hlit=b->read(5)+257;
	hdist=b->read(5)+1;
	hclen=b->read(4)+4;

	int code_len[19];

	for (i=0;i<19;i++) {
		code_len[i]=0;
	}

	//reihenfolge, in der die hclen-code-lengths im datenstrom gespeichert sind.
	int hclen_len_order[] = {16,17,18,0,8,7,9,6,10,5,11,4,12,3,13,2,14,1,15};
	for (i=0;i<hclen;i++) {
		code_len[hclen_len_order[i]]=b->read(3);
	}

	int hclen_codes[19];

	for (i=0;i<19;i++) {
		hclen_codes[i]=0;
	}

	create_code_table(code_len, hclen_codes, 7, 19);
	bin_tree *len_tree = build_tree(code_len,hclen_codes,19);

	int lit_len[286],dist_len[32]; // bitlaengen fuer literal- und distance -codelaengen
	int lit_tab[286],dist_tab[32]; // literal- und distance -codes

	decode_len(b,len_tree,lit_len,hlit);
	decode_len(b,len_tree,dist_len,hdist);
	delete len_tree;

	create_code_table(lit_len,lit_tab,15,hlit);
	create_code_table(dist_len,dist_tab,15,hdist);

	*ret_lit=build_tree(lit_len,lit_tab,hlit);
	*ret_dist=build_tree(dist_len,dist_tab,hdist);

	return 0;
}

void create_fixed_trees(bin_tree **lit, bin_tree **dist) {
	int i;
	int litcode[288],litlen[288];
	int nextcode[]={0x30, 0x190, 0x00, 0xc0};

	for (i=0;i<288;i++) {
		if (i<=143) {
			litlen[i]=8;
			litcode[i]=nextcode[0]++;
		} else if (i>=144&&i<=255) {
			litlen[i]=9;
			litcode[i]=nextcode[1]++;
		} else if (i>=256&&i<=279) {
			litlen[i]=7;
			litcode[i]=nextcode[2]++;
		} else if (i>=280&&i<=287) {
			litlen[i]=8;
			litcode[i]=nextcode[3]++;
		}
	}
	*lit=build_tree(litlen,litcode,288);
	*dist=0;
}

int decompress(buffer *buf, bin_stream *b, bin_tree *lit, bin_tree *dist) {
	int code;
	int lencode[]={3,4,5,6,7,8,9,10,11,13,15,17,19,23,
		27,31,35,43,51,59,67,83,99,115,131,163,
		195,227,258};

	// fuer den code i verwende len_extrabit[i] extrabits.
	int len_extrabit[]={0,0,0,0,0,0,0,0,1,1,1,1,2,2,2,
		2,3,3,3,3,4,4,4,4,5,5,5,5,0};

	int distcode[]={1,2,3,4,5,7,9,13,17,25,33,49,65,97,129,
		193,257,385,513,769,1025,1537,2049,3073,4097,
		6145,8193,12289,16385,24577};
	// siehe kommentar zu len_extrabits
	int dist_extrabit[]={0,0,0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,
		7,8,8,9,9,10,10,11,11,12,12,13,13};

	int l,d;

	do {
		code=get_code(b,lit);
		if (code<256) {
			buf->addbyte(code);
		}

		if (code>256) {
			code=code-257;
			l=lencode[code];
			if (len_extrabit[code]) {
				l=l+b->read(len_extrabit[code]);
			}


			// wenn datenblock "fixed codes" benutzt, verwende 5 bit, sonst den dist-tree
			if (dist==0) {
				int x,i;
				x=b->read(5);
				code=0;
				for (i=0;i<5;i++) {
					code=(code<<1)|(x&1);
					x=x>>1;
				}
			} else {
				code=get_code(b,dist);
			}
			d=distcode[code];
			if (dist_extrabit[code]) {
				d=d+b->read(dist_extrabit[code]);
			}

			buf->copy(l,d);
		}
	} while (code!=256);


	return 0;
}

int main(int argc, char **argv) {
	FILE *z;
	int BFINAL, BTYPE;
	int LEN, NLEN;
	bin_tree *lit,*dist;

	// gzip-format
	int ID1,ID2,CM,FLG,XFL,OS;
	unsigned MTIME;
 
	if (argc != 2) {
		cerr << "Syntax:\n\t" << argv[0] << " file.gz" << endl;
		return -1;
	}

	if ((z=fopen(argv[1],"r"))==NULL) {
		perror("fopen");
		return -1;
	}
	//auslesen von gzip-daten. siehe rfc1952
	ID1=fgetc(z);
	ID2=fgetc(z);
	if (ID1!=0x1f || ID2!=0x8b) {
		cerr << argv[1] << ": Datei ist kein gzip-File!"<<endl;
		fclose(z);
		return -1;
	}
	CM=fgetc(z);
	if (CM!=8) {
		cerr << "Datei ist moeglicherweise beschaedigt.\nNicht spezifiziertes Kompressionsformat.\n" << endl;
		return -1;
	}
	FLG=fgetc(z);
	fread(&MTIME,4,1,z);
	XFL=fgetc(z);
	OS=fgetc(z);


	if ((FLG&4)!=0) { // FLG.FEXTRA
		int xlen;
		xlen=fgetc(z);
		xlen=xlen|(fgetc(z)<<8);
		while(xlen--) {
			fgetc(z); // skip extra
		}
	}

	if ((FLG&8)!=0) { // FLG.FNAME
		while(fgetc(z)!='\0') { // skip filename
			continue;
		}
	}

	if ((FLG&16)!=0) { // FLG.FCOMMENT
		while (fgetc(z)!='\0') { // skip comment
			continue;
		}
	}

	if ((FLG&2)!=0) { // FLG.FHCRC
		fgetc(z); // skip crc16
		fgetc(z);
	}

	//bin_stream-klasse ermoeglicht, einzelne bits zu verarbeiten.
	bin_stream b(z);

	// buffer-klasse enthaelt den  32kb datenpuffer, der im deflate-algorithmus verwendet wird
	buffer buf;

	do {

		//header-bits
		BFINAL=b.read(1);
		BTYPE=b.read(2);

		cerr << "BFINAL="<<BFINAL<<", BTYPE="<<BTYPE<<endl;
		switch (BTYPE) {
			case 0: // unkomprimierter block
				b.flush();
				LEN=b.read(16);
				NLEN=b.read(16);
				if (LEN != (NLEN^0xffff)) {
					cerr << "Fehler im Datenstrom, len=0x" << hex << LEN << ", nlen=0x" << hex << NLEN << endl;
					exit(1);
				}
				while (LEN--) {
					buf.addbyte(b.read(8));
				}
				break;

			case 1: // "fixed codes" block
				create_fixed_trees(&lit,&dist);
				break;
			case 2: // "dynamic codes" block
				create_dynamic_trees(&b,&lit,&dist);
				break;


			case 3: // BTYPE=11(binaer) nicht definiert
				cerr << "BTYPE = 3: No valid Blocktype!" << endl;
				exit(1);
		}
		if (BTYPE) {
			decompress(&buf,&b,lit,dist);
			delete lit;
			if (dist) { // falls BTYPE==1, dist nicht definiert.
				delete dist;
			}
		}

	} while (BFINAL==0);

	buf.flush();
	fclose(z);

	return 0;
}

